'' Copyright (c) 2010 Philip C. Pilgrim
'' (See end of file for terms of use.)
''
''-----------[ Description ]----------------------------------------------------
''
'' This Propeller object monitors a serial input pin for NMEA data, extracts
'' latitude and longitude from any GPRMC sentences it encounters, and reports the
'' values extracted in various forms.
''
''-----------[ Program ]--------------------------------------------------------

CON

  STRLEN        = 80

VAR

  byte  gps_data[STRLEN + 2], lat_hemi, lon_hemi
  word  pattern, lat_deg, lat_min, lat_fmin, lon_deg, lon_min, lon_fmin
  long  t0, t1

OBJ

  io    : "basic_sio"
  re    : "regex"

PUB start (gps_port)

  ' Initializes the object, using the designated port for the GPS input.

  io.start_explicit(gps_port, -1, 4800)
  pattern := string("GPRMC\w*,[^,]*,[^,]*,($1\d+)\.($2\d+)\w*,($3N|S)\w*,($4\d+)\.($5\d+)\w*,($6E|W)")
  t1 := clkfreq * 5

PUB capture

  return \_capture
           
PRI _capture | i, n, fld

  ' Captures coordinates from GPRMC sentence. Times out if no GPRMC within five seconds.
  ' Returns true on success; false on timeout.

  t0 := cnt              
  repeat
    repeat until input == "$"     
    repeat i from 0 to 4
      gps_data[i] := input
    gps_data[5]~
    if (strcomp(@gps_data, string("GPRMC")) )
      repeat
        gps_data[i] := input
      until gps_data[i++] == "*" or i => strlen
      gps_data[i]~
      if (re.match(@gps_data, pattern, re#NOALT))
      
        n := strsize(re.field(1))
        lat_deg := dec(re.substr(re.field(1), 0, n - 2))
        lat_min := dec(re.substr(re.field(1), n - 2, 2))
        lat_fmin := dec(re.field(2))
        if ((n := strsize(re.field(2))) < 4)
          lat_fmin *= lookupz(3 - n : 10, 100, 1000, 10000)
        else
          lat_fmin /= lookupz(n - 4 : 1, 10, 100, 1000, 10000)
        lat_hemi := byte[re.field(3)]
        
        n := strsize(re.field(4))
        lon_deg := dec(re.substr(re.field(4), 0, n - 2))
        lon_min := dec(re.substr(re.field(4), n - 2, 2))
        lon_fmin := dec(re.field(5))
        if ((n := strsize(re.field(5))) < 4)
          lon_fmin *= lookupz(3 - n : 10, 100, 1000, 10000)
        else
          lon_fmin /= lookupz(n - 4 : 1, 10, 100, 1000, 10000)
        lon_hemi := byte[re.field(6)]
        return true

PUB rmc_string

  ' Returns the address of the last GPRMC sentence captured.

  return @gps_data

PUB lat_degrees

  ' Returns the whole degrees of latitude of the last GPRMC sentence captured.

  result := lat_deg
  if (lat_hemi == "S")
    -result

PUB lat_minutes

  'Returns the whole minutes of latitude of the last GPRMC sentence captured.

  result := lat_min
  if (lat_hemi == "S")
    -result

PUB lat_fminutes

  'Returns the 1/10_000ths minutes of latitude of the last GPRMC sentence captured.

  result := lat_fmin
  if (lat_hemi == "S")
    -result

PUB lat_min_0000

  'Returns the minutes part of latitude in 1/10_000ths minutes of the last GPRMC sentence captured.

  result := lat_min * 10000 + lat_fmin
  if (lat_hemi == "S")
    -result

PUB lat_in_min_0000

  'Returns the complete latitude of the last GPRMC sentence captured in 1/10_000ths minutes.

  result := lat_degrees * 60_0000 + lat_min_0000           
     
PUB lon_degrees

  ' Returns the whole degrees of longitude of the last GPRMC sentence captured.
  
  result := lon_deg
  if (lon_hemi == "W")
    -result

PUB lon_minutes

  'Returns the whole minutes of longitude of the last GPRMC sentence captured.

  result := lon_min
  if (lon_hemi == "W")
    -result

PUB lon_fminutes

  'Returns the 1/10_000ths minutes of longitude of the last GPRMC sentence captured.

  result := lon_fmin
  if (lon_hemi == "W")
    -result

PUB lon_min_0000

  'Returns the minutes part of longitude in 1/10_000ths minutes of the last GPRMC sentence captured.

  result := lon_min * 10000 + lon_fmin
  if (lon_hemi == "W")
    -result

PUB lon_in_min_0000

  'Returns the complete latitude of the last GPRMC sentence captured in 1/10_000ths minutes.

  result := lon_degrees * 60_0000 + lon_min_0000           
     
PRI dec(straddr)

  repeat strsize(straddr)
    result := result * 10 + byte[straddr++] - "0"

PRI input

  repeat until io.inbytes
    if (cnt - t0 > t1)
      abort false
  return io.in

''-----------[ TERMS OF USE ]---------------------------------------------------
''
'' Permission is hereby granted, free of charge, to any person obtaining a copy of
'' this software and associated documentation files (the "Software"), to deal in
'' the Software without restriction, including without limitation the rights to use,
'' copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
'' Software, and to permit persons to whom the Software is furnished to do so,
'' subject to the following conditions: 
''
'' The above copyright notice and this permission notice shall be included in all
'' copies or substantial portions of the Software. 
''
'' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
'' INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
'' PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
'' HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
'' OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'' SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.